==> TopdataTopFinderProSW6-v2.0.0.zip: TopdataTopFinderProSW6/src/Storefront/PageLoader/ProductDeviceListingLoader.php <==
<?php

declare(strict_types=1);

namespace Topdata\TopdataTopFinderProSW6\Storefront\PageLoader;

use Doctrine\DBAL\Connection;
use Shopware\Core\Content\Product\Events\ProductListingPreviewCriteriaEvent;
use Shopware\Core\Content\Product\ProductCollection;
use Shopware\Core\Content\Product\ProductDefinition;
use Shopware\Core\Content\Product\SalesChannel\ProductAvailableFilter;
use Shopware\Core\Content\Product\SalesChannel\ProductCloseoutFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Entity;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\DataAbstractionLayer\Search\EntitySearchResult;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\NotFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Grouping\FieldGrouping;
use Shopware\Core\Framework\DataAbstractionLayer\Search\IdSearchResult;
use Shopware\Core\Framework\Struct\ArrayEntity;
use Shopware\Core\Framework\Uuid\Uuid;
use Shopware\Core\System\SalesChannel\Entity\SalesChannelRepository;
use Shopware\Core\System\SalesChannel\SalesChannelContext;
use Shopware\Core\System\SystemConfig\SystemConfigService;
use Symfony\Contracts\EventDispatcher\EventDispatcherInterface;

/**
 * Class ProductDeviceListingLoader
 *
 * This class is responsible for loading product device listings in a sales channel context.
 */
class ProductDeviceListingLoader
{
    /**
     * @readonly
     */
    private SalesChannelRepository $productRepository;
    /**
     * @readonly
     */
    private SystemConfigService $systemConfigService;
    /**
     * @readonly
     */
    private Connection $connection;
    /**
     * @readonly
     */
    private EventDispatcherInterface $eventDispatcher;
    public function __construct(SalesChannelRepository   $productRepository, SystemConfigService      $systemConfigService, Connection               $connection, EventDispatcherInterface $eventDispatcher)
    {
        $this->productRepository = $productRepository;
        $this->systemConfigService = $systemConfigService;
        $this->connection = $connection;
        $this->eventDispatcher = $eventDispatcher;
    }

    /**
     * Loads the products of a topdataDevice based on the given criteria and sales channel context.
     *
     * @param Criteria $originalCriteria The original search criteria.
     * @param SalesChannelContext $salesChannelContext The sales channel context.
     * @return EntitySearchResult The search result containing the product listings.
     */
    public function load(Criteria $originalCriteria, SalesChannelContext $salesChannelContext): EntitySearchResult
    {
        $criteria = clone $originalCriteria;

        $this->addGrouping($criteria);
        $this->_handleAvailableStock($criteria, $salesChannelContext);

        //$context->getContext()->addState(Context::STATE_ELASTICSEARCH_AWARE);

        $ids = $this->productRepository->searchIds($criteria, $salesChannelContext);
        $aggregations = $this->productRepository->aggregate($criteria, $salesChannelContext);

        // no products found, no need to continue
        if (empty($ids->getIds())) {
            return new EntitySearchResult(
                ProductDefinition::ENTITY_NAME,
                0,
                new ProductCollection(),
                $aggregations,
                $originalCriteria,
                $salesChannelContext->getContext()
            );
        }

        $variantIds = $ids->getIds();
        $mapping = array_combine($ids->getIds(), $ids->getIds());

        if (!$this->hasOptionFilter($criteria)) {
            list($variantIds, $mapping) = $this->_resolvePreviews($ids->getIds(), $salesChannelContext);
        }

        $criteriaClonedForRead = $criteria->cloneForRead($variantIds);
        $criteriaClonedForRead->addAssociation('options.group');

        $entities = $this->productRepository->search($criteriaClonedForRead, $salesChannelContext);
        $this->_addExtensions($ids, $entities, $mapping);

        $result = new EntitySearchResult(ProductDefinition::ENTITY_NAME, $ids->getTotal(), $entities->getEntities(), $aggregations, $originalCriteria, $salesChannelContext->getContext());
        $result->addState(...$ids->getStates());

        return $result;
    }

    /**
     * Checks if the criteria has an option filter.
     *
     * @param Criteria $criteria The search criteria.
     * @return bool True if the criteria has an option filter, false otherwise.
     */
    private function hasOptionFilter(Criteria $criteria): bool
    {
        $fields = $criteria->getFilterFields();

        $fields = array_map(function (string $field) {
            return preg_replace('/^product./', '', $field);
        }, $fields);

        if (\in_array('options.id', $fields, true)) {
            return true;
        }

        if (\in_array('optionIds', $fields, true)) {
            return true;
        }

        return false;
    }

    /**
     * Adds grouping to the criteria.
     *
     * @param Criteria $criteria The search criteria.
     */
    private function addGrouping(Criteria $criteria): void
    {
        $criteria->addGroupField(new FieldGrouping('displayGroup'));

        $criteria->addFilter(
            new NotFilter(
                NotFilter::CONNECTION_AND,
                [new EqualsFilter('displayGroup', null)]
            )
        );
    }

    /**
     * Handles the available stock based on the system configuration.
     *
     * @param Criteria $criteria The search criteria.
     * @param SalesChannelContext $salesChannelContext The sales channel context.
     */
    private function _handleAvailableStock(Criteria $criteria, SalesChannelContext $salesChannelContext): void
    {
        $salesChannelId = $salesChannelContext->getSalesChannel()->getId();
        $hide = $this->systemConfigService->get('core.listing.hideCloseoutProductsWhenOutOfStock', $salesChannelId);

        if (!$hide) {
            return;
        }

        $criteria->addFilter(new ProductCloseoutFilter());
    }

    /**
     * Resolves the previews for the given IDs.
     *
     * @param array $ids The IDs to resolve previews for.
     * @param SalesChannelContext $salesChannelContext The sales channel context.
     * @return array An array containing the sorted IDs and the remapped IDs.
     */
    private function _resolvePreviews(array $ids, SalesChannelContext $salesChannelContext): array
    {
        $ids = array_combine($ids, $ids);

        $config = $this->connection->fetchAllAssociative(
            '# product-listing-loader::resolve-previews
            SELECT 
                    LOWER(HEX(child.id)) as id,
                    parent.variant_listing_config as variantListingConfig
             FROM product as child
                INNER JOIN product as parent
                    ON parent.id = child.parent_id
                    AND parent.version_id = child.version_id
             WHERE child.version_id = :version
             AND child.id IN (:ids)',
            [
                'ids'     => Uuid::fromHexToBytesList(array_values($ids)),
                'version' => Uuid::fromHexToBytes($salesChannelContext->getContext()->getVersionId()),
            ],
            ['ids' => Connection::PARAM_STR_ARRAY]
        );

        $mapping = [];

        foreach ($config as $item) {
            //{"extensions":[],"displayParent":null,"mainVariantId":null,"configuratorGroupConfig":null}
            $variantListingConfig = unserialize($item['variantListingConfig']);
            //var_dump($variantListingConfig);
            if ($variantListingConfig['mainVariantId']) {
                $mapping[$item['id']] = $variantListingConfig['mainVariantId'];
            }
        }

        // now we have a mapping for "child => main variant"
        if (empty($mapping)) {
            return [$ids, array_combine($ids, $ids)];
        }

        // filter inactive and not available variants
        $criteria = new Criteria(array_values($mapping));
        $criteria->addFilter(new ProductAvailableFilter($salesChannelContext->getSalesChannel()->getId()));
        $this->_handleAvailableStock($criteria, $salesChannelContext);

        $this->eventDispatcher->dispatch(
            new ProductListingPreviewCriteriaEvent($criteria, $salesChannelContext)
        );

        $available = $this->productRepository->searchIds($criteria, $salesChannelContext);

        $remapped = [];
        // replace existing ids with main variant id
        $sorted = [];
        foreach ($ids as $id) {
            // id has no mapped main_variant - keep old id
            if (!isset($mapping[$id])) {
                $sorted[] = $id;
                $remapped[$id] = $id;

                continue;
            }

            // get access to main variant id over the fetched config mapping
            $main = $mapping[$id];

            // main variant is configured but not active/available - keep old id
            if (!$available->has($main)) {
                $sorted[] = $id;
                $remapped[$id] = $id;

                continue;
            }

            // main variant is configured and available - add main variant id
            if (!\in_array($main, $sorted, true)) {
                $remapped[$id] = $main;
                $sorted[] = $main;
            }
        }

        return [$sorted, $remapped];
    }

    /**
     * Adds extensions to the entities based on the IDs.
     *
     * @param IdSearchResult $ids The search result containing the IDs.
     * @param EntitySearchResult $entities The search result containing the entities.
     * @param array $mapping The mapping of IDs to main variant IDs.
     */
    private function _addExtensions(IdSearchResult $ids, EntitySearchResult $entities, array $mapping): void
    {
        foreach ($ids->getExtensions() as $name => $extension) {
            $entities->addExtension($name, $extension);
        }

        foreach ($ids->getIds() as $id) {
            if (!isset($mapping[$id])) {
                continue;
            }

            // current id was mapped to another variant
            if (!$entities->has($mapping[$id])) {
                continue;
            }

            /** @var Entity $entity */
            $entity = $entities->get($mapping[$id]);

            // get access to the data of the search result
            $entity->addExtension('search', new ArrayEntity($ids->getDataOfId($id)));
        }
    }
}

==> TopdataTopFinderProSW6-v7.2.2.zip: TopdataTopFinderProSW6/src/Storefront/PageLoader/ProductDeviceListingLoader.php <==
<?php

declare(strict_types=1);

namespace Topdata\TopdataTopFinderProSW6\Storefront\PageLoader;

use Doctrine\DBAL\Connection;
use Shopware\Core\Content\Product\Events\ProductListingPreviewCriteriaEvent;
use Shopware\Core\Content\Product\ProductCollection;
use Shopware\Core\Content\Product\ProductDefinition;
use Shopware\Core\Content\Product\SalesChannel\ProductAvailableFilter;
use Shopware\Core\Content\Product\SalesChannel\ProductCloseoutFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Entity;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\DataAbstractionLayer\Search\EntitySearchResult;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\NotFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Grouping\FieldGrouping;
use Shopware\Core\Framework\DataAbstractionLayer\Search\IdSearchResult;
use Shopware\Core\Framework\Struct\ArrayEntity;
use Shopware\Core\Framework\Uuid\Uuid;
use Shopware\Core\System\SalesChannel\Entity\SalesChannelRepository;
use Shopware\Core\System\SalesChannel\SalesChannelContext;
use Shopware\Core\System\SystemConfig\SystemConfigService;
use Symfony\Contracts\EventDispatcher\EventDispatcherInterface;

/**
 * Class ProductDeviceListingLoader
 *
 * This class is responsible for loading product device listings in a sales channel context.
 */
class ProductDeviceListingLoader
{
    public function __construct(
        private readonly SalesChannelRepository   $productRepository,
        private readonly SystemConfigService      $systemConfigService,
        private readonly Connection               $connection,
        private readonly EventDispatcherInterface $eventDispatcher
    )
    {
    }

    /**
     * Loads the products of a topdataDevice based on the given criteria and sales channel context.
     *
     * @param Criteria $originalCriteria The original search criteria.
     * @param SalesChannelContext $salesChannelContext The sales channel context.
     * @return EntitySearchResult The search result containing the product listings.
     */
    public function load(Criteria $originalCriteria, SalesChannelContext $salesChannelContext): EntitySearchResult
    {
        $criteria = clone $originalCriteria;

        $this->addGrouping($criteria);
        $this->_handleAvailableStock($criteria, $salesChannelContext);

        //$context->getContext()->addState(Context::STATE_ELASTICSEARCH_AWARE);

        $ids = $this->productRepository->searchIds($criteria, $salesChannelContext);
        $aggregations = $this->productRepository->aggregate($criteria, $salesChannelContext);

        // no products found, no need to continue
        if (empty($ids->getIds())) {
            return new EntitySearchResult(
                ProductDefinition::ENTITY_NAME,
                0,
                new ProductCollection(),
                $aggregations,
                $originalCriteria,
                $salesChannelContext->getContext()
            );
        }

        $variantIds = $ids->getIds();
        $mapping = array_combine($ids->getIds(), $ids->getIds());

        if (!$this->hasOptionFilter($criteria)) {
            list($variantIds, $mapping) = $this->_resolvePreviews($ids->getIds(), $salesChannelContext);
        }

        $criteriaClonedForRead = $criteria->cloneForRead($variantIds);
        $criteriaClonedForRead->addAssociation('options.group');

        $entities = $this->productRepository->search($criteriaClonedForRead, $salesChannelContext);
        $this->_addExtensions($ids, $entities, $mapping);

        $result = new EntitySearchResult(ProductDefinition::ENTITY_NAME, $ids->getTotal(), $entities->getEntities(), $aggregations, $originalCriteria, $salesChannelContext->getContext());
        $result->addState(...$ids->getStates());

        return $result;
    }

    /**
     * Checks if the criteria has an option filter.
     *
     * @param Criteria $criteria The search criteria.
     * @return bool True if the criteria has an option filter, false otherwise.
     */
    private function hasOptionFilter(Criteria $criteria): bool
    {
        $fields = $criteria->getFilterFields();

        $fields = array_map(function (string $field) {
            return preg_replace('/^product./', '', $field);
        }, $fields);

        if (\in_array('options.id', $fields, true)) {
            return true;
        }

        if (\in_array('optionIds', $fields, true)) {
            return true;
        }

        return false;
    }

    /**
     * Adds grouping to the criteria.
     *
     * @param Criteria $criteria The search criteria.
     */
    private function addGrouping(Criteria $criteria): void
    {
        $criteria->addGroupField(new FieldGrouping('displayGroup'));

        $criteria->addFilter(
            new NotFilter(
                NotFilter::CONNECTION_AND,
                [new EqualsFilter('displayGroup', null)]
            )
        );
    }

    /**
     * Handles the available stock based on the system configuration.
     *
     * @param Criteria $criteria The search criteria.
     * @param SalesChannelContext $salesChannelContext The sales channel context.
     */
    private function _handleAvailableStock(Criteria $criteria, SalesChannelContext $salesChannelContext): void
    {
        $salesChannelId = $salesChannelContext->getSalesChannel()->getId();
        $hide = $this->systemConfigService->get('core.listing.hideCloseoutProductsWhenOutOfStock', $salesChannelId);

        if (!$hide) {
            return;
        }

        $criteria->addFilter(new ProductCloseoutFilter());
    }

    /**
     * Resolves the previews for the given IDs.
     *
     * @param array $ids The IDs to resolve previews for.
     * @param SalesChannelContext $salesChannelContext The sales channel context.
     * @return array An array containing the sorted IDs and the remapped IDs.
     */
    private function _resolvePreviews(array $ids, SalesChannelContext $salesChannelContext): array
    {
        $ids = array_combine($ids, $ids);

        $config = $this->connection->fetchAllAssociative(
            '# product-listing-loader::resolve-previews
            SELECT 
                    LOWER(HEX(child.id)) as id,
                    parent.variant_listing_config as variantListingConfig
             FROM product as child
                INNER JOIN product as parent
                    ON parent.id = child.parent_id
                    AND parent.version_id = child.version_id
             WHERE child.version_id = :version
             AND child.id IN (:ids)',
            [
                'ids'     => Uuid::fromHexToBytesList(array_values($ids)),
                'version' => Uuid::fromHexToBytes($salesChannelContext->getContext()->getVersionId()),
            ],
            ['ids' => Connection::PARAM_STR_ARRAY]
        );

        $mapping = [];

        foreach ($config as $item) {
            //{"extensions":[],"displayParent":null,"mainVariantId":null,"configuratorGroupConfig":null}
            $variantListingConfig = unserialize($item['variantListingConfig']);
            //var_dump($variantListingConfig);
            if ($variantListingConfig['mainVariantId']) {
                $mapping[$item['id']] = $variantListingConfig['mainVariantId'];
            }
        }

        // now we have a mapping for "child => main variant"
        if (empty($mapping)) {
            return [$ids, array_combine($ids, $ids)];
        }

        // filter inactive and not available variants
        $criteria = new Criteria(array_values($mapping));
        $criteria->addFilter(new ProductAvailableFilter($salesChannelContext->getSalesChannel()->getId()));
        $this->_handleAvailableStock($criteria, $salesChannelContext);

        $this->eventDispatcher->dispatch(
            new ProductListingPreviewCriteriaEvent($criteria, $salesChannelContext)
        );

        $available = $this->productRepository->searchIds($criteria, $salesChannelContext);

        $remapped = [];
        // replace existing ids with main variant id
        $sorted = [];
        foreach ($ids as $id) {
            // id has no mapped main_variant - keep old id
            if (!isset($mapping[$id])) {
                $sorted[] = $id;
                $remapped[$id] = $id;

                continue;
            }

            // get access to main variant id over the fetched config mapping
            $main = $mapping[$id];

            // main variant is configured but not active/available - keep old id
            if (!$available->has($main)) {
                $sorted[] = $id;
                $remapped[$id] = $id;

                continue;
            }

            // main variant is configured and available - add main variant id
            if (!\in_array($main, $sorted, true)) {
                $remapped[$id] = $main;
                $sorted[] = $main;
            }
        }

        return [$sorted, $remapped];
    }

    /**
     * Adds extensions to the entities based on the IDs.
     *
     * @param IdSearchResult $ids The search result containing the IDs.
     * @param EntitySearchResult $entities The search result containing the entities.
     * @param array $mapping The mapping of IDs to main variant IDs.
     */
    private function _addExtensions(IdSearchResult $ids, EntitySearchResult $entities, array $mapping): void
    {
        foreach ($ids->getExtensions() as $name => $extension) {
            $entities->addExtension($name, $extension);
        }

        foreach ($ids->getIds() as $id) {
            if (!isset($mapping[$id])) {
                continue;
            }

            // current id was mapped to another variant
            if (!$entities->has($mapping[$id])) {
                continue;
            }

            /** @var Entity $entity */
            $entity = $entities->get($mapping[$id]);

            // get access to the data of the search result
            $entity->addExtension('search', new ArrayEntity($ids->getDataOfId($id)));
        }
    }
}
